;;; permas_nas.el --- PERMAS NAS mode (editing PERMAS NASTRAN-Door files)

;; Copyright (C) 1999 Free Software Foundation, Inc.

;; Last Change: 31-Jul-15
;; Keywords: PERMAS, data, editing

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; permas-nas-mode is a major mode for editing PERMAS NASTRAN-Door files.

;;; Code:

;;;###autoload
(defun permas-nas-mode ()
  "Major mode for NASTRAN-Door input files of PERMAS."
  (interactive)
  (kill-all-local-variables) (text-mode)
  (make-local-variable 'comment-start) (setq comment-start "$")
  (make-local-variable 'tab-stop-list)
  (setq tab-stop-list '( 8 16 24 32 40 48 56 64 72 80))
  (setq mode-name "PERMAS-NAS") (setq major-mode 'permas-nas-mode)
  (substitute-key-definition 'split-line 'nas-split-line (current-local-map) global-map)
  (set (make-local-variable 'font-lock-defaults) '(permas-nas-font-lock-keywords t t))
  (if font-lock-mode (font-lock-mode 1)) ;no guess why we need this
  (run-mode-hooks 'permas-nas-mode-hook))


;; font-lock support (list of regular expressions for highlighting)
;;-----------------------------------------------------------------
(setq permas-nas-font-lock-keywords
  (list
   '("\$.*" . 'font-lock-comment-face)   ;; Comment
   '("^ *\\(cend\\|begin *bulk\\|enddata\\).*$" . 'font-lock-match-face)
   '("^ *\\(sol\\|subcase\\) *[^ \n]*"          . 'font-lock-separator-face)
   '("^ *\\(title\\|subtitle\\|label\\) *= \\([^\n]+\\)" . 'font-lock-function-name-face)
   '("^ *\\([a-z][^=]*\\)= \\([^\n]+\\)"
    (1 font-lock-type-face) (2 font-lock-variable-name-face))
   '("^[a-z][^ *,]*"                   . 'font-lock-keyword-face)
))


(defun nasdoor-document () "Document abbreviation used for NASTRAN-Door helps."
  (if (stringp permas-doc-version)
      (concat "nasdoor_" permas-doc-version) "nasdoor"))
;;
(defun nas-split-line ()
  "Break line at point and insert continuation marker and alignment."
  (interactive) (delete-horizontal-space)
  (if (save-excursion (beginning-of-line) (looking-at " *\$"))
      (insert "\n\$ ") (insert "\n\+       ")))

;;****************************************************************************
;;*****   Help on NASTRAN Keywords out of *nasdoor* documentation !  *********
;;****************************************************************************
(defun help-bulk-keyword (&optional query)
"Show help for Bulk Data keyword of actual line (info from nasdoor document).
Ask interactively for keyword if QUERY is non-nil (short cut otherwise)."
 (interactive) (call-help-with-default 'help-bulk-key 'keyword-at-point
                                        query "Bulk Data keyword"))
;
(defun help-bulk-key (key) "Display help for Bulk Data KEYword."
   (let ((actual-buffer (current-buffer)))
        (if (load-document (nasdoor-document)) (progn (goto-char (point-min))
            (search-forward "Bulk  Data  Section" nil t) ;Goto Bulk-Data-Headline
            (if (re-search-forward (concat "^  ? ? ?|+ *" key) nil t)
                (show-help-for-bulk key)
                (beep) (if (re-search-forward (concat "^    *|+ *" key) nil t)
                       (message (concat "Bulk Data Keyword \"" key
                                "\" NOT SUPPORTED/documented (search failed)"))
                       (message (concat "No such Bulk Data Keyword \""
                                         key "\" (search failed)"))))
            (set-buffer actual-buffer)))))
;
(defun show-help-for-bulk (key) "Display help for help-bulk-key function."
  (let ((pos (point))(pmax (point)) (line "Data Items")(len))
  (with-output-to-temp-buffer "*Help*"
      (print-headline-and-page (nasdoor-document) "Bulk Data" (upcase key) t)
      (if (search-forward line nil t) (setq pmax (point)))
      (princ "\n\n") (goto-char pos) (print-bulk-tex-block key)
      (while (and (< (point) pmax)
             (re-search-forward (concat "| *" key) pmax t))
             (princ "\n") (print-bulk-tex-block key))
      (goto-char pmax) (forward-line 1)
      (setq line (buffer-linestring)) (setq len (length line))
      (while (< 5 len) (princ (concat "\n" (substring line 5)))
             (forward-line 1) (setq line (buffer-linestring))
             (setq len (length line)))
      (highlight-buffer nil "*Help*" 'permas-help-mode 'auto)
      (print-help-return-message))
))
;
(defun print-bulk-tex-block (key) " "
  (let ((KEY (upcase key))
        (indent (length key))
        (pos (point))(pmax (point)))
      (princ
"<--F1--><--F2--><--F3--><--F4--><--F5--><--F6--><--F7--><--F8--><--F9-->\n")
      (print-bulk-tex-line t)
      (princ (string-to-length (make-string indent ?=) 8))
      (if (re-search-forward "^ *$" nil t) (setq pmax (point))) (goto-char pos)
      (if (re-search-forward "^ +|[^_]" pmax t)
          (progn (forward-line 1) (print-bulk-tex-line)
                 (while (re-search-forward "||||.*$" pmax t)
                        (if (re-search-forward "^ +|" pmax t)
                            (progn (princ "        ")(forward-line 1)
                                   (print-bulk-tex-line)))))
          (princ "\n"))))
;
(defun print-bulk-tex-line (&optional keyline)
"Read next 2 lines from buffer and interprete it as TeX-Bulk-Data-Line."
(let ((line1 (buffer-linestring))(line2 " ")(i 0)(field 1))
     (forward-line) (setq line2 (buffer-linestring))
     (if (not (string-match "____________________" line2))
         (setq line1 (include-subscript-line line1 line2)))
     (if keyline (progn
        (string-match " *|+ *\\([^|]+\\) *|" line1 i)
        (setq i (match-end 0)) (princ (string-to-length
        (substring line1 (match-beginning 1) (match-end 1)) 8 "L"))))
     (while (and (<= field 8) (string-match " *|+ *\\([^| ]*\\)" line1 i))
            (setq i (match-end 1)) (princ (string-to-length
            (substring line1 (match-beginning 1) i) 8 "R"))
            (setq field (+ field 1)))
     (princ "\n")))
;
(defun nas-section ()
"Determine the actual NASTRAN input section at cursor position.
Returns \"Exec\", \"Case\", \"Bulk\" or \"Unknown\" as string !"
  (save-excursion
  (let ((maxpos (min (+ (point) 10000) (point-max))))
       (beginning-of-line 1)
       (if (search-forward "cend" maxpos t) (eval "Exec")
           (if (search-forward "begin bulk" maxpos t) (eval "Case")
               (eval "Bulk")))
)))
(defun help-case-key (key) "Display help for Case Control  KEYword."
   (let ((actual-buffer (current-buffer)) pmin pmax line len
         (KEY (concat "^   ? ? ? ?|+    *" key ".*|")))
        (if (load-document (nasdoor-document)) (progn (goto-char (point-min))
            (search-forward "Case  Control  Section" nil t) ;Goto Case-Control-Headline
            (setq pmin (point))
            (if (search-forward "Bulk  Data  Section" nil t)
                (progn (setq pmax (point)) (goto-char pmin)))
            (if (re-search-forward KEY pmax t)
                (with-output-to-temp-buffer "*Help*"
                     (show-naskey-syntax "Case Control" key)
                     (print-indented-lines "Option" 5)
                     (print-indented-lines "Argument" 5)
                     (highlight-buffer nil "*Help*" 'permas-help-mode))
                (beep) (if (re-search-forward (concat "^    *|+ *" key) nil t)
                       (message (concat "Case Control Keyword \"" key
                                "\" NOT SUPPORTED/documented (search failed)"))
                       (message (concat "No such Case Control Keyword \""
                                         key "\" (search failed)"))))
            (set-buffer actual-buffer)))))
;
(defun help-exec-key (key) "Display help for Executive Control KEYword."
   (let ((actual-buffer (current-buffer)) pmin pmax line len
         (KEY (concat "^   ? ? ? ?|+    *" key ".*|")))
        (if (load-document (nasdoor-document)) (progn (goto-char (point-min))
            (search-forward "Executive  Control  Section" nil t) ;Goto Case-Control-Headline
            (setq pmin (point))
            (if (search-forward "Case  Control  Section" nil t)
                (progn (setq pmax (point)) (goto-char pmin)))
            (if (re-search-forward KEY pmax t)
                (with-output-to-temp-buffer "*Help*"
                     (show-naskey-syntax "Executive Control" key)
                     (print-indented-lines "Argument" 0)
                     (if (equal (upcase key) "SOL") (print-solseq-table))
                     (highlight-buffer nil "*Help*" 'permas-help-mode))
                (beep) (if (re-search-forward (concat "^    *|+ *" key) nil t)
                       (message (concat "Executive Control Keyword \"" key
                                "\" NOT SUPPORTED/documented (search failed)"))
                       (message (concat "No such Executive Control Keyword \""
                                         key "\" (search failed)"))))
            (set-buffer actual-buffer)))))

(defun help-global-key (key) "Display help for Executive Control KEYword."
   (let ((actual-buffer (current-buffer)) pmin pmax line len
         (KEY (concat "^   ? ? ? ?|+    *" key ".*|")))
        (if (load-document (nasdoor-document)) (progn (goto-char (point-min))
            (search-forward "Global  NASTRAN  Input" nil t) ;Goto Case-Control-Headline
            (setq pmin (point))
            (if (search-forward "Executive  Control  Section" nil t)
                (progn (setq pmax (point)) (goto-char pmin)))
            (if (re-search-forward KEY pmax t)
                (with-output-to-temp-buffer "*Help*"
                     (show-naskey-syntax "Global keyword" key)
                     (print-indented-lines "Argument" 0)
                     (highlight-buffer nil "*Help*" 'permas-help-mode))
                (beep) (message (concat "No such NASTRAN keyword \""
                                         key "\" (search failed)")))
            (set-buffer actual-buffer)))))

(defun print-solseq-table () "Special print-out for Help on NASTRAN SOL-Statement."
  (princ "\n
Table of known SOLution Sequences:
==================================")
  (if (search-forward "Solution  Sequences" nil t) (progn
      (if (search-forward " 1  " nil t) (print-indented-lines nil 13 73))
      (if (search-forward "|   " nil t) (print-indented-lines nil 12 72))
)))
(defun print-indented-lines (pattern indent &optional maxlen) " "
   (if pattern (progn (search-forward pattern nil t)
               (princ (concat "\n" (buffer-linestring))) (forward-line 1)))
   (let*((line (buffer-linestring)) (len (length line)) (last len))
        (while (< indent len)
               (setq last (if maxlen (min len maxlen) len))
               (princ (concat "\n" (substring line indent last)))
               (forward-line 1) (setq line (buffer-linestring))
               (setq len (length line)))))

(defun show-naskey-syntax (type key) ""
;   (message key-id)
  (let ((pos (point)) (indent (string-match "[^ |]" (buffer-linestring))))
       (print-headline-and-page (nasdoor-document) type  (upcase key) t)
       (princ "\n") (goto-char pos) (print-syntax-block indent t)))

(defun help-nas-keyword (&optional query)
"Show help for Bulk Data keyword of actual line (info from nasdoor document).
Ask interactively for keyword if QUERY is non-nil (short cut otherwise)."
 (interactive)
 (let ((section (nas-section))
       (key (keyword-at-point))
       (key-id (integer-at-point)))
      (cond ((string-match "^\\$" key) (help-global-key "\\[ input  \\]   $"))
            ((string-match "^PAG" key) (help-global-key key))
            ((string-match "^INC" key) (help-global-key key))
            ((string-match "^ECHOO" key) (help-global-key key))
            ((equal section "Exec") (help-exec-key key))
            ((equal section "Case") (help-case-key key))
            (t                      (help-bulk-key key))
)))

;;----------------------------------------------------------------------------
(provide 'permas_nas)

;;; permas_nas.el ends here
